#!/usr/bin/env python
# SPDX-License-Identifier: GPL-3.0-or-later
import subprocess
import statistics
import sys

bench_executable = sys.argv[1]
total_dabs = 1000
dabs_per_message = 1
runs_to_average = 50

# Generated by running `bench_multidab blendmodes`
blend_modes = [
    {"enum_name": "DP_BLEND_MODE_ERASE", "svg_name": "svg:dst-out"},
    {"enum_name": "DP_BLEND_MODE_NORMAL", "svg_name": "svg:src-over"},
    {"enum_name": "DP_BLEND_MODE_MULTIPLY", "svg_name": "svg:multiply"},
    {"enum_name": "DP_BLEND_MODE_DIVIDE", "svg_name": "krita:divide"},
    {"enum_name": "DP_BLEND_MODE_BURN", "svg_name": "svg:color-burn"},
    {"enum_name": "DP_BLEND_MODE_DODGE", "svg_name": "svg:color-dodge"},
    {"enum_name": "DP_BLEND_MODE_DARKEN", "svg_name": "svg:darken"},
    {"enum_name": "DP_BLEND_MODE_LIGHTEN", "svg_name": "svg:lighten"},
    {"enum_name": "DP_BLEND_MODE_SUBTRACT", "svg_name": "krita:subtract"},
    {"enum_name": "DP_BLEND_MODE_ADD", "svg_name": "svg:plus"},
    {"enum_name": "DP_BLEND_MODE_RECOLOR", "svg_name": "svg:src-atop"},
    {"enum_name": "DP_BLEND_MODE_BEHIND", "svg_name": "svg:dst-over"},
    {"enum_name": "DP_BLEND_MODE_COLOR_ERASE", "svg_name": "-dp-cerase"},
    {"enum_name": "DP_BLEND_MODE_SCREEN", "svg_name": "svg:screen"},
    {
        "enum_name": "DP_BLEND_MODE_NORMAL_AND_ERASER",
        "svg_name": "-dp-normal-and-eraser",
    },
    {
        "enum_name": "DP_BLEND_MODE_LUMINOSITY_SHINE_SAI",
        "svg_name": "krita:luminosity_sai",
    },
    {"enum_name": "DP_BLEND_MODE_OVERLAY", "svg_name": "svg:overlay"},
    {"enum_name": "DP_BLEND_MODE_HARD_LIGHT", "svg_name": "svg:hard-light"},
    {"enum_name": "DP_BLEND_MODE_SOFT_LIGHT", "svg_name": "svg:soft-light"},
    {"enum_name": "DP_BLEND_MODE_LINEAR_BURN", "svg_name": "krita:linear_burn"},
    {"enum_name": "DP_BLEND_MODE_LINEAR_LIGHT", "svg_name": "krita:linear light"},
    {"enum_name": "DP_BLEND_MODE_HUE", "svg_name": "svg:hue"},
    {"enum_name": "DP_BLEND_MODE_SATURATION", "svg_name": "svg:saturation"},
    {"enum_name": "DP_BLEND_MODE_LUMINOSITY", "svg_name": "svg:luminosity"},
    {"enum_name": "DP_BLEND_MODE_COLOR", "svg_name": "svg:color"},
    {"enum_name": "DP_BLEND_MODE_REPLACE", "svg_name": "-dp-replace"},
]


def run_bench(*args):
    sys.stdout.flush()
    result = subprocess.check_output([bench_executable, *map(str, args)], text=True)
    return float(result.strip())

def to_cost(results):
    return statistics.median(results)

def bench_pixel(pixel_type, alpha, blend_mode):
    results = []
    for i in range(runs_to_average):
        results.append(
            run_bench(
                pixel_type, total_dabs, dabs_per_message, alpha, blend_mode, 255, 255
            )
            / total_dabs
        )
    return to_cost(results) / (255.0 * 255.0)


def bench_classic(alpha, blend_mode):
    results = []
    for i in range(runs_to_average):
        results.append(
            run_bench(
                "classic",
                total_dabs,
                dabs_per_message,
                alpha,
                blend_mode,
                65535,
                255,
                127,
            )
            / total_dabs
        )
    return to_cost(results) / (65535.0 * 65535.0)


def bench_mypaint(lock_alpha, colorize, posterize, mode):
    results = []
    for i in range(runs_to_average):
        results.append(
            run_bench(
                "mypaint",
                total_dabs,
                dabs_per_message,
                255,
                65535,
                255,
                127,
                0,
                0,
                lock_alpha,
                colorize,
                posterize,
                mode,
            )
            / total_dabs
        )
    return to_cost(results) / (65535.0 * 65535.0)


def generate_drawpile_cost(brush_type, fn):
    sys.stdout.write(
        f"double DP_dab_cost_{brush_type}(bool indirect, int blend_mode) {{\n"
    )
    sys.stdout.write("    if (indirect) {\n")

    indirect_cost = fn(255, "svg:src-over")
    sys.stdout.write(f"        return {indirect_cost};\n")

    sys.stdout.write("    }\n")
    sys.stdout.write("    else {\n")
    sys.stdout.write("        switch (blend_mode) {\n")

    costs = [indirect_cost]
    for blend_mode in blend_modes:
        cost = fn(0, blend_mode["svg_name"])
        costs.append(cost)
        sys.stdout.write(f"        case {blend_mode['enum_name']}:\n")
        sys.stdout.write(f"            return {cost};\n")
    sys.stdout.write("        default:\n")
    sys.stdout.write('            DP_debug("Unhandled blend mode %d", blend_mode);\n')
    sys.stdout.write(f"            return {max(costs)};\n")

    sys.stdout.write("        }\n")
    sys.stdout.write("    }\n")
    sys.stdout.write("}\n\n")


def generate_pixel_cost(name, pixel_type):
    generate_drawpile_cost(
        name, lambda alpha, blend_mode: bench_pixel(pixel_type, alpha, blend_mode)
    )


def generate_classic_cost():
    generate_drawpile_cost(
        "classic", lambda alpha, blend_mode: bench_classic(alpha, blend_mode)
    )


def generate_mypaint_cost():
    sys.stdout.write(
        "double DP_dab_cost_mypaint(bool indirect, uint8_t lock_alpha, uint8_t colorize, uint8_t posterize) {\n"
    )
    sys.stdout.write("    if (indirect) {\n")
    sys.stdout.write(f"        return {bench_mypaint(0, 0, 0, 129)};\n")
    sys.stdout.write("    }\n")
    sys.stdout.write("    else {\n")
    sys.stdout.write("        double cost = 0.0;\n")
    sys.stdout.write("        if (lock_alpha != 0) {\n")
    sys.stdout.write(f"            cost += {bench_mypaint(255, 0, 0, 0)};\n")
    sys.stdout.write("        }\n")
    sys.stdout.write("        if (colorize != 0) {\n")
    sys.stdout.write(f"            cost += {bench_mypaint(0, 255, 0, 0)};\n")
    sys.stdout.write("        }\n")
    sys.stdout.write("        if (posterize != 0) {\n")
    sys.stdout.write(f"            cost += {bench_mypaint(0, 0, 255, 127)};\n")
    sys.stdout.write("        }\n")
    sys.stdout.write("        if (lock_alpha != UINT8_MAX && colorize != UINT8_MAX && posterize != UINT8_MAX) {\n")
    sys.stdout.write(f"            cost += {bench_mypaint(0, 0, 0, 0)};\n")
    sys.stdout.write("        }\n")
    sys.stdout.write("        return cost;\n")
    sys.stdout.write("    }\n")
    sys.stdout.write("}\n\n")

sys.stdout.write("// SPDX-License-Identifier: GPL-3.0-or-later\n")
sys.stdout.write("// Auto-generated by generate_dab_cost.py\n")
sys.stdout.write("#include \"dab_cost.h\"\n")
sys.stdout.write("#include <dpcommon/common.h>\n")
sys.stdout.write("#include <dpmsg/blend_mode.h>\n")
sys.stdout.write("\n")

generate_pixel_cost("pixel", "pixelround")
generate_pixel_cost("pixel_square", "pixelsquare")
generate_classic_cost()
generate_mypaint_cost()
